# ----------------------------------------------------------------------
# $Id: matrix.tcl,v 1.5 1996/04/08 19:51:46 drew Exp $
#
# File containing procedures dealing with matrices. 
#
# Author: Drew van Camp (drew@cs.toronto.edu)
#
# Copyright (c) 1995-1996 The University of Toronto.
#
# See the file "copyright" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# ----------------------------------------------------------------------

# ----------------------------------------------------------------------
# loadMatrix - loads numbers stored in a file into an already existing
# matrix. 
#
# Each line of the file will become a row in the matrix. Each element
# on the line goes into the corresponding column. There must be
# exactly as many lines in the file as rows in the matrix, and each
# line must contain exactly as many elements as the matrix has
# columns. 
#
# If the optional "transpose" argument is non-zero, rows and columns are
# inverted (in the discussion above and in the matrix itself)
#
# Returns the name of the matrix.
# ----------------------------------------------------------------------

if 0 {
proc loadMatrix { matrix file { transpose 0 } } {

    if { "$transpose" } {
	set row "inner" ;
	set col "outer" ;
    } else {
	set row "outer" ;
	set col "inner" ;
    }

    lassign [d_matrix size $matrix] limit($row) limit($col)  ;

    set idx(outer)	0 ;
    for_file values $file {
	set idx(inner)	0 ;
	foreach value $values {
	    d_matrix entryset $matrix $idx($row) $idx($col) $value ;
	    incr idx(inner) ;
	}

	if { $idx(inner) != $limit(inner) } {
	    error "expected $limit(inner) values, but got $idx(inner)"
	}
	incr idx(outer) ;
    }

    if { $idx(outer) != $limit(outer) } {
	error "expected $limit(outer) lines, but got $idx(outer)" ;
    }

    return $matrix ;
}

# ----------------------------------------------------------------------
# loadMatrixColumn - loads numbers stored in a file into an already
# existing matrix's column.
#
# Same as "loadMatrix", but you must specify the column to load the
# numbers into. File should contain one number per line. You can also
# specify "transpose" to load the numbers into a row.
#
# Returns the name of the matrix.
# ----------------------------------------------------------------------

proc loadMatrixColumn { matrix file columnIndex { transpose 0 } } {
    if { "$transpose" } {
	set row "inner" ;
	set col "outer" ;
    } else {
	set row "outer" ;
	set col "inner" ;
    }

    lassign [d_matrix size $matrix] limit($row) limit($col)  ;
    set index(inner) $columnIndex ;

    if { $index(inner) < 0 || $index(inner) >= $limit(inner) } {
	error "expected integer between 0 and [expr {$limit(inner)-1}] but got \"$index(inner)\""
    }

    set index(outer) 0 ;
    for_file value $file {
	d_matrix entryset $matrix $index($row) $index($col) $value ;
	incr index(outer) ;
    }
    if { $index(outer) != $limit(outer) } {
	error "expected $limit(outer) lines, but got $index(outer)"
    }

    return $matrix ;
}
}

# ----------------------------------------------------------------------
# pmatrix - prints a matrix out to an open file descriptor
#
# The procedure prints out the matrix one row to a line to an open
# file descriptor (by default "stdout").
# ----------------------------------------------------------------------

proc pmatrix { matrix { fileId stdout } } {
    lassign [d_matrix size $matrix] rows cols ;
    loop row 0 $rows {
	set line	{} ;
	loop col 0 $cols {
	    lappend line [d_matrix entryset $matrix $row $col] ;
	}
	puts $fileId $line
    }
}
