# ----------------------------------------------------------------------
# $Id: perms.tcl,v 1.5 1996/03/07 18:30:34 drew Exp $
#
# A procedure to run over all the files in a directory and set the
# owners and permissions properly (i.e. changes the owner of all files
# owned by the person running the script to the owner of the path, and
# sets them writable by everyone in the group (iff they're in the same
# group as the owner)).
#
# The "chown" may not work so we put it in a catch.
#
# Author: Drew van Camp (drew@cs.toronto.edu)
#
# Copyright (c) 1995-1996 The University of Toronto.
#
# See the file "copyright" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# ----------------------------------------------------------------------

proc d_perms { path } {
    file stat $path stat ;

    set uid	[id userid] ;
    set gid	[id groupid] ;

    set files [split [exec find $path -user $uid -print] "\n"] ;
    if { $gid == $stat(gid) } {
	foreach file $files {
	    if [catch {
		chmod g+w $file ;
	    } msg] {
		puts "warning: chmod: $msg" ;
	    }
	}
    }

    if { $uid != $stat(uid) } {
	foreach file $files {
	    if [catch {
		chown $stat(uid) $file ;
	    } msg] {
		puts "warning: chown: $msg" ;
	    }
	}
    }
}
 
