
# This file, and its companion documentation file ``range.tex'', were
# automatically generated from the file ``range.w'' using
# ``nuweb''. Any changes to either file will be lost the next time
# ``nuweb'' is run.
#
# See the ``copyright'' file for information on usage and
# redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# $Id: range.w,v 1.12 1996/05/03 18:45:18 drew Exp $

proc range { option range args } {
    switch $option {
        
        contains {
            if { [llength $args] != 1 } {
                error "wrong \# args: should be \"range contains range value\""
            }
            set value [lindex $args 0 ] ;
            set type  [range type $range] ;
        
            if [cequal $type enumerated] {
                return [expr { [lsearch $range $value] >= 0 }] ;
            }
        
            if { [cequal $type integer] && ![isInt $value] } {
                return 0 ;
            } elseif { [cequal $type real] && ![isFloat $value] } {
                return 0 ;
            }
        
            lassign [range limits $range] lower upper ;
        
            if { ( [isInf $lower] && ![cequal [cindex $lower 0] -] )
                 || ( ![range openlower $range] && $value == $lower)
                 || $value < $lower } {
                return 0 ;
            }
        
            if { ( [isInf $upper] && [cequal [cindex $upper 0] -] )
                 || ( ![range openupper $range] && $value == $upper)
                 || $value > $upper } {
                return 0 ;
            }
        
            return 1 ;
        }
        
        
        finite  {
            if ![lempty $args] {
                error "wrong \# args: should be \"range finite range\""
            }
            set type [range type $range] ;
        
            if { [cequal $type enumerated] } {
                return 1 ;
            }
        
            lassign [range limits $range] lower upper ;
        
            if { ( [isInf $lower] && [cequal [cindex $lower 0] -] )
                 || ( [isInf $upper] && ![cequal [cindex $upper 0] -] ) } {
                return 0 ;
            }
        
            if { [isInf $lower] || [isInf $upper] || $lower >= $upper } {
                return 1 ;
            }
        
            return [cequal $type "integer"] ;
        }
        
        
        limits {
            if ![lempty $args] {
                error "wrong \# args: should be \"range limits range\""
            }
            switch [range type $range] {
                integer {
                    set regexp "\[ \t\]*\[-+\]?(\[0-9\])+\[ \t\]*\.\.\[ \t\]*\[-+\]?(\[0-9\]\[ \t\]*)+"
                    set regexp "\[ \t\]*(\[^ \t\]*)\[ \t\]*\\.\\.\[ \t\]*(\[^ \t\]*)\[ \t\]*" ;
                }
                real {
                    set regexp "\[\\\[\\(\]\[ \t\]*(\[^ \t\]*)\[ \t\]*,\[ \t\]*(\[^ \t\]*)\[ \t\]*\[]\\)\]" ;
                }
                enumerated {
                    error "range \"$range\" doesn't have limits"
                }
             }
             regexp $regexp $range all lower upper ;
             return [list $lower $upper] ;
        }
        
        
        list    {
            if ![lempty $args] {
                error "wrong \# args: should be \"range list range\""
            }
            if ![range finite $range] {
                error "can't list an infinite range: \"$range\"" ;
            }
            set type    [range type $range] ;
        
            if { [cequal $type enumerated] } {
                set result $range ;
        
            } else {
                lassign [range limits $range] lower upper ;
        
                if { [isInf $lower] || [isInf $upper]
                     || ( $lower > $upper )
                     || ( $lower == $upper
                         && ( ![range openlower $range] || ![range openupper $range] ) ) } {
                    set result  {} ;
                } elseif { $lower == $upper } {
                    set result  $lower ;
                } else {
                    incr lower [expr { ![range openlower $range] }] ;
                    incr upper [range openupper $range] ;
                    set result {} ;
                    loop idx $lower $upper {
                        lappend result $idx ;
                    }
                }
            }
            return $result ;
        }
        
        
        openlower {
            if ![lempty $args] {
                error "wrong \# args: should be \"range openlower range\""
            }
        
            switch [range type $range] {
                "enumerated" {
                    error "enumerated ranges are neither opened or closed" ;
                }
                "integer" {
                    return 1 ;
                }
                "real" {
                    set regexp {([\[\(]).*} ;
                    regexp "^\[ \t]*$regexp" $range all brace ;
                    return [cequal $brace {[}] ;
                }
            }
        }
        
        
        openupper {
            if ![lempty $args] {
                error "wrong \# args: should be \"range openupper range\""
            }
        
            switch [range type $range] {
                "enumerated" {
                    error "enumerated ranges are neither opened or closed" ;
                }
                "integer" {
                    return 1 ;
                }
                "real" {
                    set regexp {.*([]\)])} ;
                    regexp "$regexp\[ \t]*$" $range all brace ;
                    return [cequal $brace {]}] ;
                }
            }
        }
        
        
        type {
            if ![lempty $args] {
                error "wrong \# args: should be \"range type range\""
            }
        
            set regexp "^\[ \t\]*\[\\\[\\(\](.*),(.*)\[]\\)\]\[ \t\]*\$" ;
            if [regexp $regexp $range all lower upper] {
                if { ![isFloat $lower] } {
                    error "expected floating-point limit in range \"$range\" but got \"$lower\""
                }
                if { ![isFloat $upper] } {
                    error "expected floating-point limit in range \"$range\" but got \"$upper\""
                }
                return real ;
            }
        
            set regexp "(.*)\\.\\.(.*)"
            if [regexp $regexp $range all lower upper] {
                if { ![isInt $lower] } {
                    error "expected integer limit in range \"$range\" but got \"$lower\""
                }
                if { ![isInt $upper] } {
                    error "expected integer limit in range \"$range\" but got \"$upper\""
                }
                return integer ;
            }
        
            return "enumerated" ;
        }
        
        
        default {
            error "bad option \"$option\": should be contains, finite, limits, list, openlower, openupper, or type"
        }
        
    }
}

proc isInt { value } {
    set regexp {[-+]?([0-9])+}
    expr { [regexp "^\[ \t]*$regexp\[ \t]*$" $value] || [isInf $value] }
}

proc isFloat { value } {
    set regexp {[-+]?([0-9])*(\.([0-9])*)?([eE][-+]?[0-9]+)?}
    expr { ([regexp "^\[ \t]*$regexp\[ \t]*$" $value \
                 all integer allFraction fraction exponent]
            && !([cequal $integer {}] && [cequal $fraction {}]))
           || [isInf $value] }
}

proc isInf { value } {
    set regexp {[-+]?[Ii][Nn][Ff]} ;
    regexp "^\[ \t]*$regexp\[ \t]*$" $value ;
}
