
proc saveEncoding { file { force 0 } } {
    
    set out  [wopen $file $force] ;
    
    foreach encoding [acdc_list] {
        
        set name        [lindex $encoding 1] ;
        
        
        set method      [acdc_method $name] ;
        
        
        set options {} ;
        foreach option [list passive center unit] {
            set value [acdc_options $name -$option] ;
            if ![cequal $value {}] {
                append options " $option=[list $value]"
            }
        }
        
        
        puts $out "[list $name $method] $options" ;
        
    }
    
    close $out ;
    
}

proc loadEncoding { file attrs } {
    foreach line [split [read_file -nonewline $file] "\n"] {
        
        set attr        [lvarpop line] ;
        if [catch {acdc_list $attr} list] {
            error "attribute \"$attr\" doesn't exist (file $file)" ;
        }
        set aliases     [lindex $list 0] ;
        if [lempty [union [concat [list $attr] $aliases] $attrs]] {
            error "attribute \"$attr\" exists, but isn't valid (file $file)" ;
        }
        
        
        
        set method      [lvarpop line] ;
        acdc_method $attr -- $method
        
        
        foreach key [list passive center unit] {
            acdc_options $attr -$key {} ;
        }
        
        foreach option $line {
            if ![regexp {^([^=]*)=([^=]*)$} $option all key value] {
                error "expected \"key=value\" but got \"$option\" (file $file)"
            }
            set key     [string tolower $key] ;
            acdc_options $attr -$key $value ;
        }
        
    }
}

proc genTaskInstances { cpath path fileList instList force } {
    
    global d_priv ;
    set tail          $d_priv(prototaskFile) ;
    set prototaskFile [d_getFile [file dirname $cpath]/$tail "data file"] ;
    
    
    set files [splitPrototask $prototaskFile \
                   [llength [d_info $cpath inputs]] \
                   [llength [d_info $cpath targets]] \
                   [d_info $cpath test-set-size] \
                   [d_info $cpath training-set-size] \
                   [d_info $cpath maximum-number-of-instances] \
                   [d_info $cpath test-set-selection]] ;
    set rawFiles(test)      [lvarpop files] ;
    set rawFiles(targets)   [lvarpop files] ;
    set rawFiles(training)  [lvarpop files] ;
    
    
    if [cequal $instList "all"] {
        set instList {} ;
        loop idx 0 [llength $rawFiles(training)] {
            lappend instList $idx ;
        }
    }
    
    foreach instance $instList {
        set rawTrainingFile [lindex $rawFiles(training) $instance] ;
    
        
        set attrList    [concat [d_info $cpath inputs] [d_info $cpath targets]] ;
        set indexList   {} ;
        set index       0 ;
        foreach list [eval acdc_list $attrList] {
            set type [lindex $list 0] ;
            if { [cequal $type real] || [cequal $type integer] } {
                lappend indexList $index ;
            }
            incr index ;
        }
        set matrix \
            [d_matrix create [d_info $cpath training-set-size] [llength $indexList]] ;
        loadMatrix $matrix $rawTrainingFile $indexList ;
        
        set out       [wopen $path/normalize.$instance $force] ;
        
        set col 0 ;
        foreach index $indexList {
            set attr [lindex $attrList $index] ;
            set list [d_mstats -col $col $matrix] ;
        
            puts $out [list normalization options mean $attr [lvarpop list]] ;
            puts $out [list normalization options stdDev $attr [sqrt [lvarpop list]]] ;
            puts $out [list normalization options median $attr [lvarpop list]] ;
            puts $out [list normalization options absDev $attr [lvarpop list]] ;
        
            incr col ;
        }
        
        d_matrix delete $matrix ;
        close $out ;
        
    
        set regexp \
            {^normalization options ((mean)|(stdDev)|(median)|(absDev)) (.*) (.*)}
        for_file line $path/normalize.$instance {
            if [regsub $regexp $line {acdc_options \6 -\1 \7} new] {
                eval $new ; 
            } else {
                error "unknown line in $path/normalize.$instance: \"$line\"" ;
            }
        }
    
        if { [lsearch $fileList "training"] >= 0 } {
            pwrap "  encoding instance $instance training data..." {
                set fromFileId      [open $rawTrainingFile r] ;
                set toFileId        [wopen $path/train.$instance $force] ;
                acdc_encodeFile $attrList $fromFileId $toFileId ;
                close $fromFileId ;
                close $toFileId ;
            }
        }
        if { [lsearch $fileList "testing"] >= 0 } {
            pwrap "  encoding instance $instance test inputs..." {
                set fromFileId      [open [lindex $rawFiles(test) $instance] r] ;
                set toFileId        [wopen $path/test.$instance $force] ;
                acdc_encodeFile [d_info $cpath inputs] $fromFileId $toFileId ;
                close $fromFileId ;
                close $toFileId ;
            }
        }
        if { [lsearch $fileList "targets"] >= 0 } {
            pwrap "  encoding instance $instance test targets..." {
                set fromFileId      [open [lindex $rawFiles(targets) $instance] r]
                set toFileId        [wopen $path/targets.$instance $force] ;
                acdc_encodeFile [d_info $cpath targets] $fromFileId $toFileId ;
                close $fromFileId ;
                close $toFileId ;
            }
        }
    }
    
    
    set attrList    [d_info $cpath targets] ;
    set indexList   {} ;
    set index       0 ;
    foreach list [eval acdc_list $attrList] {
        set type [lindex $list 0] ;
        if { [cequal $type real] || [cequal $type integer] } {
            lappend indexList $index ;
        }
        incr index ;
    }
    
    if { ![lempty $indexList] } {
        set matrix \
            [d_matrix create [d_info $cpath test-set-size] [llength $indexList]] ;
    
        
        set scheme      [d_info $cpath test-set-selection] ;
        if { [cequal $scheme "common"] } {
            set targetFile [lindex $rawFiles(targets) 0] ;
        } elseif { [cequal $scheme "hierarchical"] } {
            set targetFile "/tmp/alltargets.[pid]" ;
            set out        [open $targetFile w] ;
            foreach infile $rawFiles(targets) {
                set in [open $infile r] ;
                copyfile $in $out ;
                close $in ;
            }
            close $out ;
        } else {
            error "unknown test set selection scheme \"$scheme\"" ;
        }
        
        loadMatrix $matrix $targetFile $indexList ;
        if { [cequal $scheme "hierarchical"] } {
            unlink -nocomplain $targetFile ;
        }
    } elseif { [info exists matrix] } {
        unset matrix ;
    }
        
    set col 0 ;
    set variances   {} ;
    set absDevs     {} ;
    foreach index $indexList {
        set attr [lindex $attrList $index] ;
        set list [d_mstats -col $col $matrix] ;
        lappend variances   [lindex $list 1] ;
        lappend absDevs     [lindex $list 3] ;
        incr col ;
    }
    
    set out [wopen $path/Test-set-stats $force] ;
    if [info exists matrix] {
        d_matrix delete $matrix ;
        puts $out "$variances\n$absDevs"
    }
    close $out ;
    
    
    saveEncoding $path/Coding-used $force ;
    
    
    foreach key [array names rawFiles] {
        unlink -nocomplain $rawFiles($key) ;
    }
    
}

proc genTask { cpath path { force 0 } } {
    uplevel [list genTaskInstances $cpath \
                 $path "training testing targets" all $force] ;
}

proc splitPrototask { inFile numInputs numTargets 
                      numTestCases numTrainingCases maxInstances scheme} {

    
    set numCases   [exec cat $inFile | wc -l] ;
    set caseList   $numTestCases ;
    incr numCases -$numTestCases ;
    while { $numCases > 0 } {
        lappend caseList  $numTrainingCases ;
        incr numCases    -$numTrainingCases ;
    }
    set numInstances        [expr { [llength $caseList] - 1 }] ;
    
    if { $numCases != 0 } {
        error "can't divide $inFile evenly into a $numTestCases case test set and $numTrainingCases case training sets"
    }
    
    if { $numInstances > $maxInstances } {
        set numInstances $maxInstances ;
    }
    
    if { [cequal $scheme hierarchical] } {
        if { $numTestCases % $numInstances != 0 } {
            error "can't divide $numTestCases test cases into $numInstances files"
        }
        set testSetSize     [expr { $numTestCases/$numInstances }] ;
        set caseList        [eval [list lreplace $caseList 0 0] \
                                 [replicate " $testSetSize" $numInstances]] ;
    } elseif { ![cequal $scheme common] } {
        error "unknown test set selection scheme \"$scheme\"" ;
    }
    
    
    pwrap "  segmenting cases..." {
        set files [splitFile $inFile /tmp/training[pid] $caseList 1] ;
    }
    
    if { [cequal $scheme common] } {
        set file    [lvarpop files] ;
        loop idx 0 $numInstances {
            lappend uncutTestFiles  $file ;
        }
    } else {
        loop idx 0 $numInstances {
            lappend uncutTestFiles  [lvarpop files] ;
        }
    }
    loop idx 0 $numInstances {
        lappend trainingFiles       [lvarpop files] ;
    }
    unlink -nocomplain $files ;
    
    
    pwrap "  splitting test inputs and targets..." {
        foreach file $uncutTestFiles {
            set files [cutFile $file $file[pid] "$numInputs $numTargets" 1] ;
            lappend testFiles       [lvarpop files] ;
            lappend targetFiles     [lvarpop files] ;
        }
        unlink -nocomplain $uncutTestFiles ;
    }
    
    
    return [list $testFiles $targetFiles $trainingFiles] ;
    
}
