
proc cutFile { inFile outPrefix list { force 0 } } {
    set fid($inFile)    [open $inFile r] ;
    
    set outFiles    {} ;
    loop idx 0 [llength $list] {
        set outFile $outPrefix.$idx ;
        
        if [catch {wopen $outFile $force} msg] {
            
            foreach name [array names fid] {
                close $fid($name) ;
            }
            if { [info exists outFiles] && ![lempty $outFiles] } {
                unlink -nocomplain $outFiles ;
            }
            
            global errorInfo errorCode ;
            error $msg $errorInfo $errorCode ;
        }
        set fid($outFile) $msg ;
        lappend outFiles $outFile ;
        
    }
    
    
    set numColumns   0 ;
    set column      -1 ;
    foreach file $outFiles {
        set split [lvarpop list] ;
    
        incr numColumns $split ;
    
        set first($file)    [expr { $column + 1 }] ;
        set last($file)     [incr column $split] ;
    }
    
    
    set lineNum 0 ;
    while { [gets $fid($inFile) line] != -1 } {
        if { [llength $line] != $numColumns } {
            
            foreach name [array names fid] {
                close $fid($name) ;
            }
            if { [info exists outFiles] && ![lempty $outFiles] } {
                unlink -nocomplain $outFiles ;
            }
            
            error "\"$inFile\", line $lineNum has [llength $line] columns, expected $numColumns"
        }
        foreach file $outFiles {
            puts $fid($file) [lrange $line $first($file) $last($file)] ;
        }
        incr lineNum ;
    }
    
    
    foreach file $outFiles {
        close $fid($file) ;
    }
    
    close $fid($inFile) ;
    return $outFiles ;
}

proc splitFile { inFile outPrefix list { force 0 } } {
    set fid($inFile)    [open $inFile r] ;
    
    set outFiles    {} ;
    set ext         0 ;
    foreach split $list {
        set outFile $outPrefix.$ext ;
        
        if [catch {wopen $outFile $force} msg] {
            
            foreach name [array names fid] {
                close $fid($name) ;
            }
            if { [info exists outFiles] && ![lempty $outFiles] } {
                unlink -nocomplain $outFiles ;
            }
            
            global errorInfo errorCode ;
            error $msg $errorInfo $errorCode ;
        }
        set fid($outFile) $msg ;
        lappend outFiles $outFile ;
        
        
        if { [copyLines $fid($inFile) $fid($outFile) $split] != $split } {
            
            foreach name [array names fid] {
                close $fid($name) ;
            }
            if { [info exists outFiles] && ![lempty $outFiles] } {
                unlink -nocomplain $outFiles ;
            }
            
            error "$inFile has too few lines to create split $ext ($split lines)" ;
        }
        
        
        close $fid($outFile) ;
        unset fid($outFile) ;
        
        incr ext ;
    }
    
    
    if { [gets $fid($inFile) line] != -1 } {
        
        foreach name [array names fid] {
            close $fid($name) ;
        }
        if { [info exists outFiles] && ![lempty $outFiles] } {
            unlink -nocomplain $outFiles ;
        }
        
        error "$inFile has lines left over after splitting" ;
    }
    
    close $fid($inFile) ;
    return $outFiles ;
}

proc copyLines { in out numLines } {
    loop idx 0 $numLines {
        if { [gets $in line] < 0 } {
            break ;
        }
        puts $out $line ;
    }
    return $idx ;
}

proc wopen { fileName { force 0 } } {
    set mode [expr { $force ? {WRONLY CREAT TRUNC} : {WRONLY CREAT EXCL} }] ;
    open $fileName $mode ;
}

proc atPrecision { precision script } {
    global tcl_precision ;
    
    if [info exists tcl_precision] {
        set old_tcl_precision $tcl_precision ;
    }
    set tcl_precision $precision ;
    
    
    set code [catch { uplevel $script } result] ;
    
    
    if [info exists old_tcl_precision] {
        set tcl_precision $old_tcl_precision ;
    } else {
        unset tcl_precision ;
    }
    
    
    if { $code == 1 } {
        global errorInfo errorCode ;
        return -code $code -errorinfo $errorInfo -errorcode $errorCode $result
    } else {
        return -code $code $result
    }
    
}
 
proc average { list { k 0 } } {
    expr "double([join $list +])/([llength $list] - $k)" ;
}
 
proc loadMatrix { matrixHandle file args } {
    lassign [d_matrix size $matrixHandle] rows cols ;

    set row     0 ;
    if [lempty $args] {
        for_file list $file {
            set col 0 ;
            foreach value $list {
                d_matrix entryset $matrixHandle $row $col $value
                incr col ;
            }
            incr row ;
        }
    } else {
        set indices [eval concat $args] ;
        for_file list $file {
            set col 0 ;
            foreach index $indices {
                d_matrix entryset $matrixHandle $row $col \
                    [lindex $list $index] ;
                incr col ;
            }
            incr row ;
        }
    }
    if { $row != $rows } {
        error "expected $rows lines in $file but got $row" ;
    }
}
 
proc median { list } {
    
    set list   [lsort -real $list] ;
    set len    [llength $list] ;
    
    if { $len % 2 } {
        
        set n1     [expr { $len / 2 }] ;
        set median [lindex $list $n1] ;
        
    } else {
        
        set n1     [expr { $len / 2 }] ;
        set n2     [expr { $n1 - 1 }] ;
        set median [expr { 0.5*([lindex $list $n1]+[lindex $list $n2]) }] ;
        
    }
    
    return $median ;
    
}

proc square { x } {
    uplevel [list expr ($x)*($x)] ;
}

proc sum { list } {
    expr [join $list +] ;
}

proc pwrap { msg script } {
    global pwrap_silent ;
    if { [info exists pwrap_silent] && $pwrap_silent != 0 } {
        return [uplevel $script] ;
    }

    puts -nonewline stderr $msg ;
    set code [catch { uplevel $script } result] ;
    puts stderr "" ;
    if { $code == 1 } {
        global errorCode errorInfo ;
        return -code $code -errorinfo $errorInfo -errorcode $errorCode $result
    } else {
        return -code $code $result ;
    }
}

proc pwrap_percentStart { max { fileId stderr } } {
    global pc_priv ;
    global pwrap_priv ;
    
    set pc_priv(percent)    0 ;
    set pc_priv(max)        $max ;
    set pc_priv(fileId)     $fileId ;
    
    
    if { ( ![info exists pwrap_silent] || $pwrap_silent == 0 )
         && [fstat $pc_priv(fileId) tty] } {
        puts -nonewline $pc_priv(fileId) "[format {%3s %%} $pc_priv(percent)]"
    }
    
}

proc pwrap_percent { count } {
    global pc_priv ;
    global pwrap_priv ;
    
    set percent     [expr int(100*$count/$pc_priv(max))] ;
    if { $percent == $pc_priv(percent) } {
        return ;
    }
    set pc_priv(percent) $percent ;
    
    
    if { ( ![info exists pwrap_silent] || $pwrap_silent == 0 )
         && [fstat $pc_priv(fileId) tty] } {
        puts -nonewline $pc_priv(fileId) "\b\b\b\b\b[format {%3s %%} $pc_priv(percent)]" ;
    }
    
}

proc pwrap_percentEnd {} {
    global pc_priv ;
    global pwrap_priv ;
    
    if { ( ![info exists pwrap_silent] || $pwrap_silent == 0 )
         && [fstat $pc_priv(fileId) tty] } {
        puts -nonewline $pc_priv(fileId) "\b\b\b\b\b[format {%5s} {}]" ;
        puts -nonewline $pc_priv(fileId) "\b\b\b\b\b" ;
    }
    
}

proc pwrap_for_file { msg varName fileName script } {
    
    global pwrap_silent ;
    if { [info exists pwrap_silent] && $pwrap_silent != 0 } {
        return [uplevel for_file $varName $fileName $script] ;
    }
     
    pwrap $msg {
        upvar $varName line ;
        
        set lineNum 0 ;
        pwrap_percentStart [exec cat $fileName | wc -l] ;
        
        
        for_file line $fileName {
            uplevel $script ;
            pwrap_percent [incr lineNum] ;
        }
        
        
        pwrap_percentEnd ;
        
    }
}

proc lorder { list orderList } {
    set result {} ;
    foreach index $orderList {
        lappend result [lindex $list $index] ;
    }
    return $result ;
}
